# Copyright 2022 PONIO TEAM. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

#! /usr/bin/env python

import itertools
import json

import sympy as sp
import numpy as np
import matplotlib.pyplot as plt

import ponio.rungekutta as rk
import ponio.order as order

print(order)

x, y = np.linspace(-5, 2, 500), np.linspace(-3.5, 3.5, 500)
X, Y = np.meshgrid(x, y)
Z = X + 1j*Y

raw_rk44 = """{
    "label": "RK (4,4)",
    "A": [
        [
            "0",
            "0",
            "0",
            "0"
        ],
        [
            "1/2",
            "0",
            "0",
            "0"
        ],
        [
            "0",
            "1/2",
            "0",
            "0"
        ],
        [
            "0",
            "0",
            "1",
            "0"
        ]
    ],
    "b": [
        "1/6",
        "1/3",
        "1/3",
        "1/6"
    ],
    "c": [
        "0",
        "1/2",
        "1/2",
        "1"
    ]
}
"""
dict_rk44 = json.loads(raw_rk44)

rk44 = rk.runge_kutta(dict_rk44['label'], dict_rk44['A'], dict_rk44['b'])
z = rk44.z_var

for stage_i in rk44.scheme(latex=False):
    print(f"{stage_i.lhs} = {stage_i.rhs}")

R = sp.lambdify(z, rk44.stability_function()(z), 'numpy')
nR = np.abs(R(Z))

plt.contourf(X, Y, nR, [0., 1.0], colors="C0", alpha=0.25)
plt.contour(X, Y, nR, [1.0], colors="C0")

plt.title(f"Stability domain of {rk44.label}")
plt.xlabel("$\\Re(z)$")
plt.ylabel("$\\Im(z)$")
plt.axhline(0, 1, 0, linestyle="--", color="grey", linewidth=1)
plt.axvline(0, 1, 0, linestyle="--", color="grey", linewidth=1)

plt.show()

print("order: ", order.order(rk44))
print("stage order: ", order.stage_order(rk44))

lrk44 = rk.lawson_runge_kutta(
    "L"+dict_rk44['label'], dict_rk44['A'], dict_rk44['b'])
zl, zn = lrk44.z_var

for stage_i in lrk44.scheme(latex=False):
    print(f"{stage_i.lhs} = {stage_i.rhs}")

R = sp.lambdify((zl, zn), lrk44.stability_function()(zl, zn), 'numpy')

zl_values = [0., -1.0, 1j]

for i, nzl in enumerate(zl_values):
    nR = np.abs(R(nzl, Z))
    plt.contourf(X, Y, nR, [0., 1.0], colors=f"C{i}", alpha=0.25)
    plt.contour(X, Y, nR, [1.0], colors=f"C{i}")
    plt.plot([0], [0], "-", color=f"C{i}", label=f"$z_\\ell = {nzl}$")

plt.title(f"Stability domain of {lrk44.label}")
plt.xlabel("$\\Re(z_n)$")
plt.ylabel("$\\Im(z_n)$")
plt.axhline(0, 1, 0, linestyle="--", color="grey", linewidth=1)
plt.axvline(0, 1, 0, linestyle="--", color="grey", linewidth=1)

plt.legend()
plt.show()

print("order: ", order.order(lrk44))
print("stage order: ", order.stage_order(lrk44))


raw_sw = """{
    "label": "Strehmel Weiner",
    "A": [
        [
            "0",
            "0",
            "0",
            "0"
        ],
        [
            "1/2 phi_1_2",
            "0",
            "0",
            "0"
        ],
        [
            "1/2 phi_1_3 - 1/2 phi_2_3",
            "1/2 phi_2_3",
            "0",
            "0"
        ],
        [
            "phi_1_4 - 2 phi_2_4",
            "-2 phi_2_4",
            "4 phi_2_4",
            "0"
        ]
    ],
    "b": [
        "phi_1 - 3 phi_2 + 4 phi_3",
        "0",
        "4 phi_2 - 8 phi_3",
        "- phi_2 + 4 phi_3"
    ],
    "c": [
        "0",
        "1/2",
        "1/2",
        "1"
    ],
    "tag": "expRK"
}"""
dict_sw = json.loads(raw_sw)

c = sp.Matrix(dict_sw['c'])

dummy_x = sp.Dummy("x")
local_dict = {
    f"phi_{i}": rk.phi(i)(dummy_x)
    for i in range(0, len(c)+1)
}
local_dict.update({
    f"phi_{i}_{j}": rk.phi(i, j, c)(dummy_x)
    for i, j in itertools.product(range(0, len(c)+1), range(1, len(c)+1))
})

dict_sw['A'] = sp.Lambda(dummy_x, sp.Matrix([
    [sp.parse_expr(str(aij), transformations='all',
                   local_dict=local_dict) for aij in ai]
    for ai in dict_sw['A']
]))
dict_sw['b'] = sp.Lambda(dummy_x, sp.Matrix([
    sp.parse_expr(str(bi), transformations='all', local_dict=local_dict)
    for bi in dict_sw['b']
]))

sw = rk.exponential_runge_kutta(
    dict_sw['label'], dict_sw['A'], dict_sw['b'], dict_sw['c'])

zl, zn = sw.z_var
R = sp.lambdify((zl, zn), sw.stability_function()(zl, zn).doit(), 'numpy')

zl_values = [1e-3, -1.0, 3j]

for i, nzl in enumerate(zl_values):
    nR = np.abs(R(nzl, Z))
    plt.contourf(X, Y, nR, [0., 1.0], colors=f"C{i}", alpha=0.25)
    plt.contour(X, Y, nR, [1.0], colors=f"C{i}")
    plt.plot([0], [0], "-", color=f"C{i}", label=f"$z_\\ell = {nzl}$")

plt.title(f"Stability domain of {sw.label}")
plt.xlabel("$\\Re(z_n)$")
plt.ylabel("$\\Im(z_n)$")
plt.axhline(0, 1, 0, linestyle="--", color="grey", linewidth=1)
plt.axvline(0, 1, 0, linestyle="--", color="grey", linewidth=1)

plt.legend()
plt.show()

print("order: ", order.order(sw))
print("stage order: ", order.stage_order(sw))

for stage_i in sw.scheme(latex=False):
    print(f"{stage_i.lhs} = {stage_i.rhs}")
