# Copyright 2022 PONIO TEAM. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

#! /usr/bin/env python

import sympy as sp
import jinja2

import itertools

import ponio.butcher_tableau
import ponio.filter
import ponio.details


def repr_latex_generator(vphantom: str = ""):
    """
      returns a function that display a Buthcer tableau
      with a vphantom element in each line
    """

    env = jinja2.Environment(loader=jinja2.FileSystemLoader("ponio/templates"))
    butcher_tpl = env.get_template("butcher.jinja2.tex")

    def repr_latex(self, A: sp.Matrix, b: sp.Matrix, c: sp.Matrix, b2: (sp.Matrix | None) = None) -> str:
        shape_of_array = "{{ c | {} }}".format("c"*A.shape[1])

        return butcher_tpl.render(
            vphantom=vphantom,
            shape_of_array=shape_of_array,
            cA=[[sp.latex(xij) for xij in xi]
                for xi in c.row_join(A).tolist()],
            b=map(sp.latex, b.T.tolist()[0]),
            b2=map(sp.latex, b2.T.tolist()[0]) if b2 is not None else None
        )

    return repr_latex


def idify(label: str) -> str:
    id_label = label.lower()
    replacements = [(" ", "_")]
    deletements = "(),-/'"

    for old, new in replacements:
        id_label = id_label.replace(old, new)
    for char in deletements:
        id_label = id_label.replace(char, "")

    return id_label


class butcher_tableau(ponio.butcher_tableau.butcher_tableau_storage):
    r"""
      store only a Butcher tableau:

      $$
        \begin{array}{c|c}
          c & A \\
          \hline
            & b^\top
        \end{array}
      $$

      and store embedded stage if needed
    """

    repr_latex_tableau = repr_latex_generator(vphantom="")

    def _init_c(cls, c, A) -> sp.Matrix:
        if c is None:
            return sp.Matrix([sum(ai) for ai in A.tolist()])
        return sp.Matrix(c)

    def __init__(self, label: str, A, b, c=None, b2=None):
        """
          * `label`: a string with the name of current Butcher tableau
          * `A`: a list of list or a `sympy.Matrix` object
          * `b`: a list or a `sympy.Matrix` object
          * `c`: a list or a `sympy.Matrix` object or `None`, if it is `None` is set to the sum of lines of `A`
          * `b2`: a list or a `sympy.Matrix` object or `None`, if it is `None` the method is not embedded
        """

        super().__init__(
            sp.Matrix(A),
            sp.Matrix(b),
            self._init_c(c, sp.Matrix(A)),
            sp.Matrix(b2) if b2 is not None else None
        )
        self.label = label

    @property
    def is_explicit(self) -> bool:
        return ponio.filter.is_explicit(self)

    @property
    def is_dirk(self) -> bool:
        return ponio.filter.is_dirk(self)

    @property
    def is_implicit(self) -> bool:
        return ponio.filter.is_implicit(self)

    @property
    def is_fully_implicit(self) -> bool:
        return ponio.filter.is_fully_implicit(self)

    @property
    def is_approached_method(self) -> bool:
        return ponio.filter.is_approached_tableau(self)

    @property
    def is_embedded(self) -> bool:
        return ponio.filter.is_embedded(self)

    @property
    def is_stiffly_accurate(self) -> bool:
        return ponio.filter.is_stiffly_accurate(self)

    @property
    def id(self) -> str:
        """ label as a potential name of file or function """
        return idify(self.label)

    def __iter__(self):
        """ to convert object into dictionary """
        yield 'label', self.label
        yield 'A', ponio.details.list_stringify_matrix(self.A)
        yield 'b', ponio.details.list_stringify_vector(self.b)
        yield 'c', ponio.details.list_stringify_vector(self.c)

        if self.is_embedded:
            yield 'b2', ponio.details.list_stringify_vector(self.b2)


class butcher_imex:
    def __init__(self, label: str, implicit: ponio.butcher_tableau.butcher_tableau_storage, explicit: ponio.butcher_tableau.butcher_tableau_storage):
        is_valid = implicit.m_shape == explicit.m_shape and implicit.v_shape == explicit.v_shape
        if not is_valid:
            raise ValueError(
                "given Butcher tableaus are not compatible with an IMEX method")

        if not ponio.filter.is_implicit(implicit):
            raise ValueError("implicit tableau is not an implicit method")

        if not ponio.filter.is_explicit(explicit):
            raise ValueError("explicit tableau is not an explicit method")

        self.label = label
        self.implicit = implicit
        self.explicit = explicit

    def __iter__(self):
        """ to convert object into dictionary """

        yield 'label', self.label
        yield 'tableaus', [dict(self.implicit), dict(self.explicit)]

    @property
    def tableaus(self) -> list[ponio.butcher_tableau.butcher_tableau_storage]:
        return [self.implicit, self.explicit]

    def __getitem__(self, key):
        return self.tableaus[key]

    def __len__(self) -> int:
        return len(self.tableaus)

    @property
    def is_embedded(self) -> bool:
        return ponio.filter.is_embedded(self.implicit) and ponio.filter.is_embedded(self.explicit)

    @property
    def is_approached_method(self) -> bool:
        return ponio.filter.is_approached_tableau(self.implicit) or ponio.filter.is_approached_tableau(self.explicit)

    @property
    def is_stiffly_accurate(self) -> bool:
        return ponio.filter.is_stiffly_accurate(self.implicit) and ponio.filter.is_stiffly_accurate(self.explicit)

    @property
    def m_shape(self) -> tuple[int, int]:
        return self.implicit.m_shape

    @property
    def v_shape(self) -> tuple[int, int]:
        return self.implicit.v_shape

    @property
    def n_stages(self) -> int:
        """ number of stages of the Butcher tableau method """
        return self.v_shape[0]

    @property
    def id(self) -> str:
        """ label as a potential name of file or function """
        return idify(self.label)

    def _repr_rows(self) -> list[str]:
        rows_implicit = self.implicit._repr_rows()
        rows_explicit = self.explicit._repr_rows()

        return [f"{ri}    {re}" for ri, re in zip(rows_implicit, rows_explicit)]

    def __repr__(self) -> str:
        return "\n".join(self._repr_rows())

    def _latex(self, printer) -> str:
        vphantom = r"\vphantom{\frac{\sqrt{0}}{\sqrt{0}}}"
        self.implicit.vphantom = vphantom
        self.explicit.vphantom = vphantom

        return r"\qquad".join([printer._print(rk_i) for rk_i in self.tableaus])

    def _repr_latex_(self, **kwargs) -> str:
        return f"${sp.latex(self, **kwargs)}$"


class butcher_tuple:
    r"""
      store a tuple of Butcher tableau:

      $$
        \begin{array}{c|c}
          c & A \\
          \hline
            & b^\top
        \end{array}
        \quad
        \begin{array}{c|c}
          \tilde{c} & \tilde{A} \\
          \hline
            & \tilde{b}^\top
        \end{array}
        \quad
        \dots
      $$

      and store embedded stages if needed
    """

    def __init__(self, label: str, *butcher_tableaus):
        ref_m_shape = butcher_tableaus[0].m_shape
        ref_v_shape = butcher_tableaus[0].v_shape
        is_valid = all([
            tab_i.m_shape == ref_m_shape and tab_i.v_shape == ref_v_shape
            for tab_i in butcher_tableaus
        ])

        if not is_valid:
            raise ValueError(
                "given Butcher tableaus are not compatible with an additive Runge-Kutta method")
        self.label = label
        self.tableaus = list(butcher_tableaus)

    def __iter__(self):
        """ to convert object into dictionary """

        yield 'label', self.label
        yield 'tableaus', [dict(tab_i) for tab_i in self.tableaus]

    def _repr_latex_(self, **kwargs) -> str:
        return f"${r"\quad".join([self.repr_latex_tuple_tableau(tab_i.A, tab_i.b, tab_i.c, tab_i.b2) for tab_i in self.tableaus])}$"

    def __getitem__(self, key):
        return self.tableaus[key]

    def __setitem__(self, key, value):
        self.tableaus[key] = value

    def __len__(self):
        return len(self.tableaus)

    @property
    def is_embedded(self) -> bool:
        return all([ponio.filter.is_embedded(tab_i) for tab_i in self.tableaus])

    @property
    def is_approached_method(self) -> bool:
        return any([ponio.filter.is_approached_tableau(tab_i) for tab_i in self.tableaus])

    @property
    def is_stiffly_accurate(self) -> bool:
        return all([ponio.filter.is_stiffly_accurate(tab_i) for tab_i in self.tableaus])

    @property
    def m_shape(self) -> tuple[int, int]:
        return self.tableaus[0].m_shape

    @property
    def v_shape(self) -> tuple[int, int]:
        return self.tableaus[0].v_shape

    @property
    def n_stages(self) -> int:
        """ number of stages of the Butcher tableau method """
        return self.v_shape[0]

    @property
    def id(self) -> str:
        """ label as a potential name of file or function """
        return idify(self.label)

    def _repr_rows(self) -> list[str]:
        rows = [tab_i._repr_rows() for tab_i in self.tableaus]

        return [f"    ".join(row_i) for row_i in zip(*rows)]

    def __repr__(self) -> str:
        return "\n".join(self._repr_rows())

    def _latex(self, printer) -> str:
        vphantom = r"\vphantom{\frac{\sqrt{0}}{\sqrt{0}}}"
        for tab_i in self.tableaus:
            tab_i.vphantom = vphantom

        return r"\qquad".join([printer._print(tab_i) for tab_i in self.tableaus])

    def _repr_latex_(self, **kwargs) -> str:
        return f"${sp.latex(self, **kwargs)}$"
