# Copyright 2022 PONIO TEAM. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

#! /usr/bin/env python

import dataclasses

import sympy
import jinja2

import ponio.details


@dataclasses.dataclass(repr=False)
class butcher_tableau_storage:
    """ storage of a Butcher tableau """
    A: sympy.Matrix
    b: sympy.Matrix
    c: sympy.Matrix
    b2: (sympy.Matrix | None) = None
    vphantom: str = dataclasses.field(repr=False, default="")

    @property
    def m_shape(self) -> tuple[int, int]:
        """ shape of the matrix A in Butcher tableau """
        return self.A.shape

    @property
    def v_shape(self) -> tuple[int, int]:
        """ shape of a vector b or c in Butcher tableau """
        return self.b.shape

    @property
    def n_stages(self) -> int:
        """ number of stages of the Butcher tableau method """
        return self.v_shape[0]

    def __iter__(self):
        """ to convert object into dictionary """
        yield 'A', ponio.details.list_stringify_matrix(self.A)
        yield 'b', ponio.details.list_stringify_vector(self.b)
        yield 'c', ponio.details.list_stringify_vector(self.c)

        if self.is_embedded:
            yield 'b2', ponio.details.list_stringify_vector(self.b2)

    def _repr_rows(self) -> list[str]:
        col_c_len = max([len(str(ci)) for ci in self.c])
        col_A_len = max([len(str(aij)) for aij in self.A])
        col_b_len = max([len(str(bi)) for bi in self.b])
        col_Ab_len = max(col_A_len, col_b_len)

        if self.b2 is not None:
            col_b2_len = max([len(str(bi)) for bi in self.b2])
            col_Ab_len = max(col_Ab_len, col_b2_len)

        rows = []
        for i, ci in enumerate(self.c):
            row_ci_Aij = f"{str(ci):^{col_c_len}} │ {' '.join(
                [f'{str(aij):^{col_Ab_len}}' for aij in self.A.row(i)])}"
            rows.append(row_ci_Aij)

        row_hline = f"{'─'*col_c_len}─┼{'─'*((col_Ab_len+1)*self.n_stages)}"
        rows.append(row_hline)

        row_bi = f"{
            ' '*col_c_len} │ {' '.join([f'{str(bi):^{col_Ab_len}}' for bi in self.b])}"
        rows.append(row_bi)

        if self.b2 is not None:
            row_hline = f"{'─'*col_c_len}─┼{'─' *
                                            ((col_Ab_len+1)*self.n_stages)}"
            rows.append(row_hline)

            row_b2i = f"{
                ' '*col_c_len} │ {' '.join([f'{str(bi):^{col_Ab_len}}' for bi in self.b2])}"
            rows.append(row_b2i)

        return rows

    def __repr__(self) -> str:
        return "\n".join(self._repr_rows())

    def _latex(self, printer) -> str:
        env = jinja2.Environment(
            loader=jinja2.FileSystemLoader("ponio/templates"))
        butcher_tpl = env.get_template("butcher.jinja2.tex")
        shape_of_array = "{{ c | {} }}".format("c"*self.A.shape[1])

        latex_butcher = butcher_tpl.render(
            vphantom=self.vphantom,
            shape_of_array=shape_of_array,
            cA=ponio.details.list_stringify_matrix(
                self.c.row_join(self.A), printer._print),
            b=ponio.details.list_stringify_vector(self.b, printer._print),
            b2=ponio.details.list_stringify_vector(
                self.b2, printer._print) if self.b2 is not None else None
        )
        return latex_butcher

    def _repr_latex_(self, **kwargs) -> str:
        return f"${sympy.latex(self, **kwargs)}$"
