# Copyright 2022 PONIO TEAM. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

#! /usr/bin/env python

import sympy as sp
import numpy as np

from ponio.butcher_tableau import butcher_tableau_storage

# test only matrix properties


def is_lower_matrix(M: sp.Matrix) -> bool:
    return sum([
        sp.Abs(M[i, j]) for i, j in zip(*np.triu_indices(M.cols, 1))
    ]) == 0


def is_strictly_lower_matrix(M: sp.Matrix) -> bool:
    return sum([
        sp.Abs(M[i, j]) for i, j in zip(*np.triu_indices(M.cols))
    ]) == 0


def is_approached_matrix(M: sp.Matrix) -> bool:
    """ is true if any of element of the matrix is a float """
    return any([type(xi) is sp.Float for xi in M])

# filter for kind of methods


def is_explicit(tableau: butcher_tableau_storage) -> bool:
    """ checks if the Butcher tableau is a tableau of an explicit method """
    return is_strictly_lower_matrix(tableau.A)


def is_dirk(tableau: butcher_tableau_storage) -> bool:
    """ checks if the Butcher tableau is a tableau of a DIRK method """
    return is_lower_matrix(tableau.A) and not is_strictly_lower_matrix(tableau.A)


def is_implicit(tableau: butcher_tableau_storage) -> bool:
    """ checks if the Butcher tableau is a tableau of an implicit method """
    return not is_explicit(tableau)


def is_fully_implicit(tableau: butcher_tableau_storage) -> bool:
    """ checks if the Butcher tableau is a tableau of a fully implicit method """
    return not is_explicit(tableau) and not is_dirk(tableau)

# filter with Butcher tableau properties


def is_approached_tableau(tableau: butcher_tableau_storage) -> bool:
    """ check if a composant of Butcher tableau is a float """
    return is_approached_matrix(tableau.A) or is_approached_matrix(tableau.b) or is_approached_matrix(tableau.c)


def is_embedded(tableau: butcher_tableau_storage) -> bool:
    """ check if the Butcher tableau is a tableau of an embedded method (contains an extra line `b2`) """
    return hasattr(tableau, 'b2') and tableau.b2 is not None


def is_stiffly_accurate(tableau: butcher_tableau_storage) -> bool:
    """ check if the Butcher tableau is a tableau of a stiffly accurate method """
    return tableau.b == tableau.A[-1, :].T

# filter with stability function of the Runge-Kutta method


def is_A_stable(rk) -> bool:
    R = rk.stability_function()
    z, = R.signature

    limit = sp.Limit(R(z), z, sp.I*sp.oo).doit().evalf()
    return limit <= 1


def is_L_stable(rk) -> bool:
    if is_A_stable(rk):
        if not is_stiffly_accurate(rk):
            return False

        if rk.A.det() == 0:
            tol = 1e-5
            # Formula 10 in https://doi.org/10.1137/080713562
            return ((rk.A[1:-1, 1:-1]).inv())[-1, :]*rk.A[1:-1, 0] < tol

        return True

    return False
