# Copyright 2022 PONIO TEAM. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

#! /usr/bin/env python

import sympy as sp
import numpy as np

from ponio.runge_kutta import runge_kutta


def order(rk: runge_kutta) -> int:
    r""" compute order from the order star of the Runge-Kutta method

    The number of petals of the order star near to $0$ is the order of the method plus one. To estimate it, compute the number of change of sign of $\mathcal{A}(r e^{i\theta}) - 1$, with $\mathcal{A}$ the function of the order star, $r = 0.25$ a small radius near to $0$, and $\theta\in[0, 2\pi]$.
    """
    A = rk.order_star_function()

    z = A.signature[-1]

    zeros = np.zeros(len(A.signature) - 1)
    expr_A = A(*zeros, z).simplify()

    nA = sp.lambdify(z, expr_A, 'numpy')

    num_order = 0

    r = 0.25
    theta = np.linspace(0., 2.*np.pi, 100)
    Z = r*np.exp(1j*theta)

    circle_A = np.abs(nA(Z)) - 1
    for i, zi in enumerate(circle_A):
        if zi*circle_A[i-1] < 0:
            num_order += 1

    return int(num_order/2) - 1


def stage_order(rk: runge_kutta) -> int:
    r""" compute stage order of a Runge-Kutta method with the following equation:

    The stage order $r$ is the highest number such:
    $$
        \forall i,\ \sum_j a_{ij}c_j^{s-1} = \frac{1}{s}c_i^s,\quad s=1,\dots,r
    $$
    """

    s = 1

    # hack to set all free symbols in matrix A (in case of expRK method) to 0
    z = set().union(*[aij.free_symbols for aij in rk.A])
    zeros = np.zeros(len(z))

    def equalities(s):
        return [
            float(sum([aij.subs(zip(z, zeros)).doit() * cj**(s - 1)
                       for aij, cj in zip(rk.A.row(i), rk.c)]) - (rk.c[i]**s)/s) < 1e-5
            for i in range(0, rk.n_stages)
        ]

    while all(equalities(s)):
        s += 1

    return s - 1


def error_constant(rk):
    b = rk.b
    A = rk.A
    p = order(rk)
    e = sp.ones(*b.shape)

    return sp.Abs((b.T*A**p*e)[0] - sp.Rational(1, sp.factorial(p+1)))


def index_1_order_condition(rk):
    raise NotImplementedError

    edirk = all([a0j == 0 for a0j in rk.A[0, :]])
    if edirk:
        omega = rk.A[1:, :-1].inv()
    else:
        omega = rk.A.inv()
    # only for IMEX...
    # not implemented yet


def is_absolute_monotonic(r, rk):
    # Definition in Section 2.3 https://doi.org/10.1016/j.cam.2014.05.011
    # If found a way to do this with exact values with SymPy for RK, not for additive RK
    assert r >= 0, "The absolute monotonicity is defined only for positive values of the radius"

    e = sp.ones(*b.shape)

    aux_matrix = sp.eye(*(rk.b).n_stages()) + r*(rk.A)

    if aux_matrix.det() == 0:
        return False

    inv_matrix = aux_matrix.inv()

    return (inv_matrix*(rk.A) >= 0).all() and (inv_matrix*e >= 0).all()


def radius_absolute_monotonicity(rk):
    is_absolute_monotonic_at_r = True
    r = 0
    while (is_absolute_monotonic_at_r):
        is_absolute_monotonic_at_r = is_absolute_monotonic(r, rk)
        r = r + 1e-5

    return r
