# Copyright 2022 PONIO TEAM. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

#! /usr/bin/env python

import json
import itertools

import sympy as sp
import numpy as np

import ponio.butcher
import ponio.filter


class runge_kutta(ponio.butcher.butcher_tableau):
    @classmethod
    def from_file(cls, filename):
        with open(filename) as input_file:
            data = json.load(input_file)
            label = data['label']
            A = data['A']
            b = data['b']
            c = data['c']
            b2 = data['b2'] if 'b2' in data else None
            return cls(label=label, A=A, b=b, c=c, b2=None)

    def __init__(self, label, A, b, c=None, b2=None) -> None:
        super().__init__(label, A, b, c, b2)

        self._R = None

    @property
    def z_var(self):
        return tuple([sp.Symbol("z")])

    def internal_stability_expr(self):
        r""" the internal stability expression :

        $$
          ( I - zA )^-1 1\!\!1
        $$
        """
        I = sp.eye(*self.m_shape)     # identity matrix
        ones = sp.ones(*self.v_shape)  # vector of ones

        z, *_ = self.z_var
        tmp = (I - z*self.A).inv()

        if self.is_approached_method:
            tmp = tmp.evalf(chop=True)

        return tmp*ones

    def internal_stability_functions(self):
        r""" the internal stability function :

        $$
          R: z \mapsto ( I - zA )^-1 1\!\!1
        $$
        """
        return [sp.Lambda(self.z_var, expr.expand().simplify().collect(self.z_var[0])) for expr in self.internal_stability_expr()]

    def stability_expr(self, embedded=False):
        r""" the stability expression :

        $$
          1 + z b^\top( I - zA )^-1 1\!\!1
        $$
        """

        z, *_ = self.z_var

        def R_generator(b, internal_stab):
            return 1 + (z*(b.T*internal_stab)[0]).expand()

        if embedded:
            return R_generator(self.b2, self.internal_stability_expr())

        return R_generator(self.b, self.internal_stability_expr())

    def stability_function(self, embedded=False):
        r""" the stability function :

        $$
          R: z \mapsto 1 + z b^\top( I - zA )^-1 1\!\!1
        $$
        """
        if embedded:
            return sp.Lambda(self.z_var, self.stability_expr(embedded))

        if self._R is None:
            self._R = sp.Lambda(self.z_var, self.stability_expr())
        return self._R

    def order_star_expr(self, kind=1, embedded=False):
        r""" the first kind order star expression :

        $$
          \frac{R(z)}{e^z}
        $$

        the second kind order star expression :

        $$
          R(z) - e^z
        $$

        ## See also

        * `stability_function`
        """
        if kind == 1:
            return sp.exp(-sum(self.z_var))*self.stability_expr(embedded)
        elif kind == 2:
            return self.stability_expr(embedded) - sp.exp(sum(self.z_var))
        else:
            raise ValueError(
                f"kind (kind={kind}) argument should be equals to 1 or 2")

    def order_star_function(self, kind=1, embedded=False):
        r""" the first kind order star function :

        $$
          \mathcal{A}:z \mapsto \frac{R(z)}{e^z}
        $$

        the second kind order star function :

        $$
          \mathcal{A}:z \mapsto R(z) - e^z
        $$

        ## See also

        * `stability_function`
        """
        return sp.Lambda(self.z_var, self.order_star_expr(kind, embedded))

    def relative_error_expr(self, embedded=False):
        r""" the relative error expression :

        $$
          left|\frac{R(z)}{e^z} - 1\right|
        $$

        ## See also

        * `stability_function`
        """
        return sp.Abs(self.stability_expr(embedded)/sp.exp(sum(self.z_var)) - 1)

    def relative_error_function(self, embedded=False):
        r""" the relative error function :

        $$
          \mathcal{e}: z \mapsto left|\frac{R(z)}{e^z} - 1\right|
        $$

        ## See also

        * `stability_function`
        """
        return sp.Lambda(self.z_var, self.relative_error_expr(embedded))

    def _lambdify(self, func):
        return sp.lambdify(self.z_var, func(*self.z_var).doit(), 'numpy')

    def lambdify_internal_stability_functions(self):
        return [self._lambdify(fi) for fi in self.internal_stability_functions()]

    def lambdify_stability_function(self, embedded=False):
        return self._lambdify(self.stability_function(embedded))

    def lambdify_order_star_function(self, kind=1, embedded=False):
        return self._lambdify(self.order_star_function(kind, embedded))

    def lambdify_relative_error_function(self, embedded=False):
        return self._lambdify(self.relative_error_function(embedded))

    def order(self, embedded=False) -> int:
        nA = self.lambdify_order_star_function(kind=1, embedded=embedded)

        num_order = 0

        r = 0.25
        theta = np.linspace(0., 2.*np.pi, 100)
        Z = r*np.exp(1j*theta)

        circle_A = np.abs(nA(Z)) - 1
        for i, zi in enumerate(circle_A):
            if zi*circle_A[i-1] < 0:
                num_order += 1

        return int(num_order//2) - 1

    def stage_order(self) -> int:
        s = 1

        def _equality(s: int, i: int):
            return sp.Eq(
                sum([aij * cj**(s-1)
                    for aij, cj in zip(self.A.row(i), self.c)]),
                self.c[i]**s*sp.Rational(1, s),
                evaluate=False
            )

        def _test_equality(s: int, i: int):
            eq = _equality(s, i)
            return float(eq.lhs - eq.rhs) < 1e-5

        def _equalities(s: int):
            return [_test_equality(s, i) for i in range(0, self.n_stages)]

        while all(_equalities(s)):
            s += 1

        return s - 1

    @property
    def is_A_stable(self) -> bool:
        z, = self.z_var
        R = self.stability_expr()
        limit = sp.Limit(sp.Abs(R), z, sp.I*sp.oo).doit().evalf()
        return limit <= 1

    @property
    def is_L_stable(self) -> bool:
        if self.is_A_stable:
            if self.is_stiffly_accurate:
                if self.A.det() != 0:
                    return True
                else:
                    hat_A = self.A[1:, 1:]
                    if hat_A.det() != 0:
                        return sp.Abs((hat_A.inv()[-1, :] * self.A[1:, 0])[0]) < 1e-10

            z, = self.z_var
            R = self.stability_expr()
            limit = sp.Limit(sp.Abs(R), z, -sp.oo).doit().evalf()

            return sp.Abs(limit) < 1e-10

        return False

    def error_constant(self, embedded: bool = False) -> float:
        p = self.order(embedded)
        e = sp.ones(*self.v_shape)

        expr = (self.b.T*self.A**p*e)[0] - sp.Rational(1, sp.factorial(p+1))
        return sp.Abs(expr)

    def scheme(self, latex: bool = True):
        """ returns scheme as a list of equations """
        if latex:
            ks_tpl = "k_{{1:{}}}"
            un_tmp = r"u^n u^{n+1} \tilde{u}^{n+1}"
            t_tmp = r"t^n \Delta\ t"
        else:
            ks_tpl = "k1:{}"
            un_tmp = "un unp1 ubnp1"
            t_tmp = "tn dt"

        def _dot_product(a, b):
            return sum([ai*bi for ai, bi in zip(a, b)])

        ks = sp.symbols(ks_tpl.format(self.n_stages + 1))
        un, unp1, ubnp1 = sp.symbols(un_tmp)
        tn, dt = sp.symbols(t_tmp)

        f = sp.Function("f", nargs=2)

        stages = []
        for i in range(0, self.n_stages):
            rhs = _dot_product(self.A.row(i), ks)
            stages.append(sp.Eq(
                ks[i],
                f(tn + self.c[i]*dt, un + dt*rhs)
            ))

        # last stage
        rhs = _dot_product(self.b, ks)
        stages.append(sp.Eq(
            unp1,
            un + dt*rhs
        ))

        # if embedded
        if self.is_embedded:
            rhs = _dot_product(self.b2, ks)
            stages.append(sp.Eq(
                ubnp1,
                un + dt*rhs
            ))

        return stages


class lawson_runge_kutta(runge_kutta):
    r"""
    Lawson method, defines by its underlying Runge-Kutta method:

    $$
      \begin{array}{c|c}
        c & A \\
        \hline
          & b^\top
      \end{array}
    $$

    To solve ODE $\dot{u} = Lu + N(t, u)$ the scheme is:

    $$
      \begin{aligned}
        k_i &= e^{-c_i\Delta t L} N(t^n + c_i\Delta t, u^n + \Delta t \sum_j a_{ij}k_j)\\
        u^{n+1} &= u^n + \Delta t \sum_i b_i k_i
      \end{aligned}
    $$
    """

    def __init__(self, label, A, b, c=None, b2=None) -> None:
        super().__init__(label, A, b, c, b2)

    @property
    def z_var(self) -> tuple[sp.Symbol, sp.Symbol]:
        return sp.symbols("z_n z_l")

    def stability_expr(self, embedded=False):
        z = super().z_var
        zn, zl = self.z_var

        return sp.exp(zl)*super().stability_expr(embedded).subs(z, zn)

    def scheme(self, latex=True):
        """ returns scheme as a list of equations """
        if latex:
            ks_tpl = "k_{{1:{}}}"
            un_tmp = r"u^n u^{n+1} \tilde{u}^{n+1}"
            t_tmp = r"t^n \Delta\ t"
        else:
            ks_tpl = "k1:{}"
            un_tmp = "un unp1 ubnp1"
            t_tmp = "tn dt"

        def dot_product(a, b):
            return sp.MatAdd(*[ai*bi for ai, bi in zip(a, b)])

        # because we should fix a shape
        vec_shape = (3, 1)
        mat_shape = (3, 3)

        ks = [sp.MatrixSymbol(ks_tpl.format(i), *vec_shape)
              for i in range(1, self.n_stages + 1)]
        un, unp1, ubnp1 = [sp.MatrixSymbol(
            ui, *vec_shape) for ui in un_tmp.split(" ")]
        tn, dt = sp.symbols(t_tmp)

        L = sp.MatrixSymbol("L", *mat_shape)

        def mexp(expr):
            if type(expr) is sp.ZeroMatrix:
                return 1
            if latex:
                expr_str = r"e^{{ {} }}".format(sp.latex(expr.simplify()))
            else:
                expr_str = r"exp({})".format(
                    str(expr.simplify().evalf()).replace("1.0*", ""))
            return sp.MatrixSymbol(expr_str, *expr.shape)

        def N(t, u):
            if latex:
                expr_str = r"N\left({}, {}\right)".format(
                    sp.latex(t), sp.latex(u.simplify()))
            else:
                expr_str = r"N({}, {})".format(str(t), str(u.simplify()))
            return sp.MatrixSymbol(expr_str, *u.shape)

        stages = []
        for i in range(0, self.n_stages):
            rhs = dot_product(self.A.row(i), ks)
            stages.append(sp.Eq(
                ks[i],
                mexp(-self.c[i]*dt*L) * N(tn + self.c[i] *
                                          dt, mexp(self.c[i]*dt*L)*(un + dt*rhs))
            ))

        # last stage
        rhs = dot_product(self.b, ks)
        stages.append(sp.Eq(
            unp1,
            mexp(dt*L) * (un + dt*rhs)
        ))

        # if embedded
        if self.is_embedded:
            rhs = dot_product(self.b2, ks)
            stages.append(sp.Eq(
                ubnp1,
                mexp(dt*L) * (un + dt*rhs)
            ))

        return stages


def phi(i, j=None, c=None):
    r"""
    phi functions factory

    $$
      \varphi_i : z \mapsto \frac{e^z - \sum_{k=0}^i \frac{1}{k!}z^k}{z^i}
    $$

    with notation $\vaphi_i = \varphi_i(\Delta t L)$, and $\varphi_{i,j} = \varphi_i(c_j\Delta t L)$.
    """
    ij = j is not None and c is not None  # set True if j and c are not None (so compute phi_{ij})

    classname = f"phi_{i}" if not ij else f"phi_{i}_{j}"

    @classmethod
    def phi_i_eval(cls, z):
        pass

    def phi_i_doit(self, deep=False, **hints):
        z = self.cj*self.args[0]
        if i == 0:
            return sp.exp(z)
        elif z.is_zero:
            return sp.Rational(1, sp.factorial(i))
        else:
            return ((phi(i-1)(z) - phi(i-1)(0))/z).simplify().expand().simplify()

    def phi_i_latex(self, printer):
        return f"\\varphi_{{{i}}}"
        # return str( type(self).mro()[0] ).replace(" ","\ ")

    def phi_ij_latex(self, printer):
        return f"\\varphi_{{{i},{j}}}"

    return type(
        classname,
        (sp.Function,),
        {
            'cj': 1 if not ij else c[j-1],
            'eval':   phi_i_eval,
            'doit':   phi_i_doit,
            '_latex': phi_i_latex if not ij else phi_ij_latex
        }
    )


class exponential_runge_kutta(runge_kutta):
    r"""
    Exponential Runge-Kutta method, defines by its Butcher tableau:

    $$
      \begin{array}{c|c}
        c & A \\
        \hline
          & b^\top
      \end{array}
    $$

    To solve ODE $\dot{u} = Lu + N(t, u)$ the scheme is:

    $$
      \begin{aligned}
        u^{(i)} &= u^n u^n + \Delta t \sum_j a_{ij}(\Delta t L)\left( N(t^n + c_j\Delta t, u^{(j)}) + Lu^n \right) \\
        u^{n+1} &= u^n + \Delta t \sum_i b_i(\Delta t L)\left( N(t^n + c_i\Delta t, u^{(i)}) + Lu^n \right)
      \end{aligned}
    $$
    """

    # TODO: savoir quand est-ce qu'on évalue A et b
    # A et b doivent rester des matrices de lambda car on ne sait pas comment elles vont être évaluée (a_ij(z_l) ou a_ij(dt*L))
    # -> avoir une fonction d'évaluation de A, b et b2 en zl ou dt*L
    def __init__(self, label, A, b, c, b2=None) -> None:
        x = sp.Dummy('x')

        tmp_c = sp.Matrix(c)

        funcs_phi = {
            f"phi_{i}": phi(i)(x)
            for i in range(0, len(tmp_c)+1)
        }
        funcs_phi.update({
            f"phi_{i}_{j}": phi(i, j, tmp_c)(x)
            for i, j in itertools.product(range(0, len(tmp_c)+1), range(1, len(tmp_c)+1))
        })

        tmp_A = sp.Lambda(x, sp.Matrix([
            [
                sp.parse_expr(str(aij), transformations='all',
                              local_dict=funcs_phi)
                for aij in ai
            ]
            for ai in A
        ]))
        tmp_b = sp.Lambda(x, sp.Matrix([
            sp.parse_expr(str(bi), transformations='all',
                          local_dict=funcs_phi)
            for bi in b
        ]))
        tmp_b2 = sp.Lambda(x, sp.Matrix([
            sp.parse_expr(str(bi), transformations='all',
                          local_dict=funcs_phi)
            for bi in b2
        ])) if b2 is not None else lambda _: None

        zl, _ = self.z_var
        super().__init__(label, tmp_A(zl), tmp_b(zl), tmp_c, tmp_b2(zl))

    @property
    def z_var(self) -> tuple[sp.Symbol, sp.Symbol]:
        r""" returns dummy variables $z_\ell$ and $z_n$ for stability function """
        return sp.symbols("z_l z_n")

    def stability_function(self, embedded=False):
        r""" the stability function

        $$
            R(z_\ell, z_n) = 1 + z_n b^\top ( I - z_n A )^{-1} \cdot (I - z_\ell A) 1\!\!1 + z_\ell b^\top 1\!\!1
        $$

        By default `z` is obtain by `self.z_var` which is the tuple $(z_\ell, z_n)$ for Lawson method
        """
        def stab_function(A, b, approached_method=False):
            zl, zn = self.z_var

            I = sp.eye(*A.shape)
            ones = sp.ones(*b.shape)

            tmp = (I - zn*A).inv()

            if approached_method:
                tmp = tmp.evalf(chop=True)

            R = 1 + zn * (b.T*(tmp * (I + zl*A) * ones))[0] + zl*(b.T*ones)[0]

            return sp.Lambda((zl, zn), R)

        if embedded:
            return stab_function(self.A, self.b2, approached_method=self.is_approached_method)

        if self._R is None:
            self._R = stab_function(
                self.A, self.b, approached_method=self.is_approached_method)
        return self._R

    def internal_stability_functions(self):
        I = sp.eye(*self.A.shape)     # identity matrix
        ones = sp.ones(*self.b.shape)  # vector of ones
        zl, zn = self.z_var
        tmp = (I - zn*self.A).inv()

        if self.is_approached_method:
            tmp = tmp.evalf(chop=True)
        tmp = tmp*(I + zl*self.A)*ones

        return [sp.Lambda((zl, zn), tmp_i.expand().simplify().collect(zn)) for tmp_i in tmp]

    def scheme(self, latex=True):
        """ returns scheme as a list of equations """
        if latex:
            ui_tpl = "u^{{({})}}"
            un_tmp = r"u^n u^{n+1} \tilde{u}^{n+1}"
            t_tmp = r"t^n \Delta\ t"
        else:
            ui_tpl = "u{}"
            un_tmp = "un unp1 ubnp1"
            t_tmp = "tn dt"

        def dot_product(a, b, l, u, dt):
            # , start=sp.zeros(*a.shape))
            return sum([ai*(bi + l*u) for ai, bi in zip(a, b)])

        vec_shape = (3, 1)
        mat_shape = (3, 3)
        # us = [sp.MatrixSymbol(ui_tpl.format(i), *vec_shape)
        #       for i in range(1, self.n_stages + 1)]
        # un, unp1, ubnp1 = [sp.MatrixSymbol(
        #     ui, *vec_shape) for ui in un_tmp.split(" ")]

        us = [sp.symbols(ui_tpl.format(i))
              for i in range(1, self.n_stages + 1)]
        un, unp1, ubnp1 = sp.symbols(un_tmp)
        tn, dt = sp.symbols(t_tmp)

        N = sp.Function("N", nargs=2)
        L = sp.Symbol("L")  # sp.MatrixSymbol("L", *mat_shape)

        zl, _ = self.z_var
        A_dtL = sp.Lambda(zl, self.A)(dt*L).doit()
        b_dtL = sp.Lambda(zl, self.b)(dt*L).doit()

        Nui = [N(tn + ci*dt, ui) for ci, ui in zip(self.c, us)]

        stages = []
        for i in range(0, self.n_stages):
            rhs = dot_product(A_dtL.row(i), Nui, L, un, dt)
            stages.append(sp.Eq(
                us[i], un + dt*rhs
            ))

        # last stage
        rhs = dot_product(b_dtL, Nui, L, un, dt)
        stages.append(sp.Eq(
            unp1, un + dt*rhs
        ))

        # if embedded
        if self.is_embedded:
            rhs = dot_product(self.b2, Nui, L, un, dt)
            stages.append(sp.Eq(
                ubnp1, un + dt*rhs
            ))

        return stages


class imex_runge_kutta(ponio.butcher.butcher_imex):
    def __init__(self, label: str, implicit, explicit):
        super().__init__(label, implicit, explicit)

    @property
    def z_var(self):
        return sp.symbols("z_i z_e")

    def internal_stability_expr(self):
        r""" the internal stability expression :

        $$
          ( I - z_i A^{[i]} - z_e A^{[e]} )^-1 1\!\!1
        $$
        """
        I = sp.eye(*self.m_shape)     # identity matrix
        ones = sp.ones(*self.v_shape)  # vector of ones

        zi, ze = self.z_var
        tmp = (I - zi*self.implicit.A - ze*self.explicit.A).inv()

        if self.is_approached_method:
            tmp = tmp.evalf(chop=True)

        return tmp*ones

    def internal_stability_functions(self):
        r""" the internal stability function :

        $$
          R: z_i, z_e \mapsto ( I - z_i A^{[i]} - z_e A^{[e]} )^-1 1\!\!1
        $$
        """
        return [sp.Lambda(self.z_var, expr.expand().simplify().collect(self.z_var[0])) for expr in self.internal_stability_expr()]

    def stability_expr(self, embedded=False):
        r""" the stability expression :

        $$
          1 + (z_i b^{[i]} + z_e b^{[e]})^\top( I - z_i A^{[i]} - z_e A^{[e]} )^-1 1\!\!1
        $$
        """

        zi, ze = self.z_var

        def R_generator(bi, be, internal_stab):
            return 1 + ((zi*bi + ze*be).T*internal_stab)[0]

        if embedded:
            return R_generator(self.implicit.b2, self.explicit.b2, self.internal_stability_expr())

        return R_generator(self.implicit.b, self.explicit.b, self.internal_stability_expr())

    def stability_function(self, embedded=False):
        r""" the stability function :

        $$
          R: z_i, z_e \mapsto 1 + (z_i b^{[i]} + z_e b^{[e]})^\top( I - z_i A^{[i]} - z_e A^{[e]} )^-1 1\!\!1
        $$
        """
        if embedded:
            return sp.Lambda(self.z_var, self.stability_expr(embedded))

        return sp.Lambda(self.z_var, self.stability_expr())

    def order_star_expr(self, kind=1, embedded=False):
        r""" the first kind order star expression :

        $$
          \frac{R(z_i, z_e)}{e^{z_i+z_e}}
        $$

        the second kind order star expression :

        $$
          R(z_i, z_e) - e^{z_i + z_e}
        $$

        ## See also

        * `stability_function`
        """
        if kind == 1:
            return sp.exp(-sum(self.z_var))*self.stability_function(embedded)(*self.z_var)
        elif kind == 2:
            return self.stability_function(embedded)(*self.z_var) - sp.exp(sum(self.z_var))
        else:
            raise ValueError(
                f"kind (kind={kind}) argument should be equals to 1 or 2")

    def order_star_function(self, kind=1, embedded=False):
        r""" the first kind order star function :

        $$
          \mathcal{A}:z_i, z_e \mapsto \frac{R(z_i, z_e)}{e^{z_i + z_e}}
        $$

        the second kind order star function :

        $$
          \mathcal{A}:z_i, z_e \mapsto R(z_i, z_e) - e^{z_i + z_e}
        $$

        ## See also

        * `stability_function`
        """
        return sp.Lambda(self.z_var, self.order_star_expr(kind, embedded))

    def relative_error_expr(self, embedded=False):
        r""" the relative error expression :

        $$
          left|\frac{R(z_i, z_e)}{e^{z_i+z_e}} - 1\right|
        $$

        ## See also

        * `stability_function`
        """
        return sp.Abs(self.stability_expr(embedded)/sp.exp(sum(self.z_var)) - 1)

    def relative_error_function(self, embedded=False):
        r""" the relative error function :

        $$
          \mathcal{e}: z_i, z_e \mapsto left|\frac{R(z_i, z_e)}{e^{z_i+z_e}} - 1\right|
        $$

        ## See also

        * `stability_function`
        """
        return sp.Lambda(self.z_var, self.relative_error_expr(embedded))

    def _lambdify(self, func):
        return sp.lambdify(self.z_var, func(*self.z_var).doit(), 'numpy')

    def lambdify_internal_stability_functions(self):
        return [self._lambdify(fi) for fi in self.internal_stability_functions()]

    def lambdify_stability_function(self, embedded=False):
        return self._lambdify(self.stability_function(embedded))

    def lambdify_order_star_function(self, kind=1, embedded=False):
        return self._lambdify(self.order_star_function(kind, embedded))

    def lambdify_relative_error_function(self, embedded=False):
        return self._lambdify(self.relative_error_function(embedded))

    @property
    def is_A_stable(self) -> tuple[bool]:
        R = self.stability_function()
        zi, ze = R.signature

        limit_i = sp.Limit(R(zi, 0), zi, sp.I*sp.oo).doit().evalf()
        limit_e = sp.Limit(R(0, ze), ze, sp.I*sp.oo).doit().evalf()

        implicit_is_A_stable = limit_i <= 1
        explicit_is_A_stable = limit_e <= 1

        return (implicit_is_A_stable, explicit_is_A_stable)

    @property
    def is_L_stable(self) -> tuple[bool]:
        R = self.stability_function()
        zi, ze = R.signature

        def is_L_stable_composante(A_stab, tab, R, z) -> bool:
            if A_stab:
                if tab.A.det() != 0:
                    return tab.is_stiffly_accurate

                limit = sp.Limit(sp.Abs(R), z, -sp.oo).doit().evalf()
                return sp.Abs(limit) < 1e-5
            return False

        A_stability = self.is_A_stable

        implicit_is_L_stable = is_L_stable_composante(
            A_stability[0], self.implicit, R(zi, 0), zi)
        explicit_is_L_stable = is_L_stable_composante(
            A_stability[1], self.explicit, R(0, ze), ze)

        return (implicit_is_L_stable, explicit_is_L_stable)


class additive_runge_kutta(ponio.butcher.butcher_tuple):
    def __init__(self, label: str, *butcher_tableaus):
        super().__init__(label, *butcher_tableaus)

        self._R = None

    @property
    def z_var(self):
        return sp.symbols(f"z_:{len(self)}")

    def internal_stability_expr(self):
        r""" the internal stability expression :

        $$
          ( I - \sum_i z_i A^{[i]} )^-1 1\!\!1
        $$
        """
        I = sp.eye(*self.m_shape)     # identity matrix
        ones = sp.ones(*self.v_shape)  # vector of ones
        zeros = sp.zeros(*self.m_shape)  # zeros matrix

        tmp = (I - sum([
            z_i*tab_i.A
            for z_i, tab_i in zip(self.z_var, self.tableaus)
        ], start=zeros)).inv()

        if self.is_approached_method:
            tmp = tmp.evalf(chop=True)

        return tmp*ones

    def internal_stability_functions(self):
        r""" the internal stability function :

        $$
          R: z_i, z_e \mapsto ( I - \sum_i z_i A^{[i]} )^-1 1\!\!1
        $$
        """
        return [sp.Lambda(self.z_var, expr.expand().simplify().collect(self.z_var[0])) for expr in self.internal_stability_expr()]

    def stability_expr(self, embedded=False):
        r""" the stability expression :

        $$
          1 + (\sum_i z_i b^{[i]})^\top( I - \sum_i z_i A^{[i]} )^-1 1\!\!1
        $$
        """

        def R_generator(b_list, internal_stab):
            zeros = sp.zeros(*self.v_shape)
            return 1 + (sum([zi*bi for zi, bi in zip(self.z_var, b_list)], start=zeros).T*internal_stab)[0]

        if embedded:
            return R_generator([rki.b2 for rki in self.tableaus], self.internal_stability_expr())

        return R_generator([rki.b for rki in self.tableaus], self.internal_stability_expr())

    def stability_function(self, embedded=False):
        r""" the stability function :

        $$
          R: z \mapsto 1 + (\sum_i z_i b^{[i]})^\top( I - \sum_i z_i A^{[i]} )^-1 1\!\!1
        $$
        """
        if embedded:
            return sp.Lambda(self.z_var, self.stability_expr(embedded))

        if self._R is None:
            self._R = sp.Lambda(self.z_var, self.stability_expr())
        return self._R

    def order_star_expr(self, kind=1, embedded=False):
        r""" the first kind order star expression :

        $$
          \frac{R(z)}{e^{\sum z_i}}
        $$

        the second kind order star expression :

        $$
          R(z) - e^{\sum z_i}
        $$

        ## See also

        * `stability_function`
        """
        if kind == 1:
            return sp.exp(-sum(self.z_var))*self.stability_function(embedded)(*self.z_var)
        elif kind == 2:
            return self.stability_function(embedded)(*self.z_var) - sp.exp(sum(self.z_var))
        else:
            raise ValueError(
                f"kind (kind={kind}) argument should be equals to 1 or 2")

    def order_star_function(self, kind=1, embedded=False):
        r""" the first kind order star function :

        $$
          \mathcal{A}:z \mapsto \frac{R(z)}{e^{\sum z_i}}
        $$

        the second kind order star function :

        $$
          \mathcal{A}:z \mapsto R(z) - e^{\sum z_i}
        $$

        ## See also

        * `stability_function`
        """
        return sp.Lambda(self.z_var, self.order_star_expr(kind, embedded))

    def relative_error_expr(self, embedded=False):
        r""" the relative error expression :

        $$
          left|\frac{R(z)}{e^{\sum z_i}} - 1\right|
        $$

        ## See also

        * `stability_function`
        """
        return sp.Abs(self.stability_expr(embedded)/sp.exp(sum(self.z_var)) - 1)

    def relative_error_function(self, embedded=False):
        r""" the relative error function :

        $$
          \mathcal{e}: z_i, z_e \mapsto left|\frac{R(z)}{e^{\sum z_i}} - 1\right|
        $$

        ## See also

        * `stability_function`
        """
        return sp.Lambda(self.z_var, self.relative_error_expr(embedded))

    def _lambdify(self, func):
        return sp.lambdify(self.z_var, func(*self.z_var).doit(), 'numpy')

    def lambdify_internal_stability_functions(self):
        return [self._lambdify(fi) for fi in self.internal_stability_functions()]

    def lambdify_stability_function(self, embedded=False):
        return self._lambdify(self.stability_function(embedded))

    def lambdify_order_star_function(self, kind=1, embedded=False):
        return self._lambdify(self.order_star_function(kind, embedded))

    def lambdify_relative_error_function(self, embedded=False):
        return self._lambdify(self.relative_error_function(embedded))

    @property
    def is_A_stable(self) -> tuple[bool]:
        R = self.stability_function()
        z = R.signature

        def zi_vec(i: int) -> sp.Matrix:
            z_vec = sp.zeros(len(z), 1)
            z_vec[i] = z[i]
            return z_vec

        limits = [sp.Limit(R(*zi_vec(i)), zi, sp.I*sp.oo).doit().evalf()
                  for i, zi in enumerate(z)]

        return tuple(li <= 1 for li in limits)

    @property
    def is_L_stable(self) -> tuple[bool]:
        R = self.stability_function()
        z = R.signature

        def zi_vec(i: int) -> sp.Matrix:
            z_vec = sp.zeros(len(z), 1)
            z_vec[i] = z[i]
            return z_vec

        def is_L_stable_composante(A_stab, tab, R, z) -> bool:
            if A_stab:
                if tab.A.det() != 0:
                    return tab.is_stiffly_accurate

                limit = sp.Limit(sp.Abs(R), z, -sp.oo).doit().evalf()
                return sp.Abs(limit) < 1e-5
            return False

        A_stability = self.is_A_stable
        return tuple(
            is_L_stable_composante(A_stability_i, tab_i, R(zi_vec(i)), zi)
            for i, (A_stability_i, tab_i, zi) in enumerate(zip(A_stability, self.tableaus, z))
        )
